#include <cstdio>
#include <string>
#include <cstdarg>

#include "bdecode.hh"

autoptr<CAtom> BDecode(const std::string& input, unsigned& pos)
{
    #define GetCh() (pos>=input.size() ? EOF : input[pos++])
    #define Ungetc() { --pos; }
    
    int ch;
    switch(ch = GetCh())
    {
        case 'i':
        {
            uint_least64_t intval = 0;
            while((ch = GetCh()) != 'e')
            {
                if(ch == EOF) break;
                intval = intval* 10 + (ch-'0');
            }
            return new CAtomInt(intval);
        }
        case 'l': //list
        {
            autoptr<CAtomList> list = new CAtomList;
            while((ch = GetCh()) != 'e')
            {
                if(ch == EOF) break;
                Ungetc();
                list->addItem(BDecode(input, pos));
            }
            return &*list;
        }
        case 'd': //dict
        {
            autoptr<CAtomDict> list = new CAtomDict;
            while((ch = GetCh()) != 'e')
            {
                if(ch == EOF) break;
                Ungetc();
                
                autoptr<CAtom> key = BDecode(input, pos);
                if(!key) throw "Dictonary key should have been a string!";
                const CAtomString* str = key->get_str();
                if(!str) throw "Dictonary key should have been a string!";

                list->setItem(str->getValue(), BDecode(input, pos));
            }
            return &*list;
        }
        case '0': case '1': case '2': case '3':
        case '4': case '5': case '6': case '7':
        case '8': case '9':
        {
            unsigned len = 0;
            while(ch != ':' && ch != EOF)
            {
                len = len * 10 + (ch - '0');
                ch = GetCh();
            }
            if(ch != EOF)
            {
                unsigned oldpos = pos;
                pos += len;
                return new CAtomString( std::string(&input[oldpos], &input[pos]) );
            }
            break;
        }
        default:
            throw "Invalid character found";
    }
    return 0;
}

autoptr<CAtom> BDecode(const std::string& input)
{
    unsigned pos = 0;
    try {
        autoptr<CAtom> resp = BDecode(input, pos);
        if(pos == input.size())
            return resp;
    }
    catch(const char*)
    {
    }
    return 0;
}
